#! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
# 
# Licensed under the HARK license, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# https://www.hark.jp/notice/HARK_License_Agreement.pdf

import paho.mqtt.client as mqtt

import traceback
from harkmw.logger import logger

import os
import signal
import threading
import pickle
from copy import copy

import harkmw


class SubProcess():
    """起動された子プロセスを管理するクラス
    """

    def __init__(self, daemon, process_name):
        """起動された子プロセスの情報を初期化する

        Parameters
        ----------
        daemon : Daemon
            子プロセスを起動したデーモン
        process_name : str
            子プロセスを識別する名前
        """
        self.daemon = daemon
        self.process_name = process_name
        self.child_pid = None

    def get_status(self):
        """子プロセスの状態を取得する
        """
        if self.child_pid:
            return 'ok'
        else:
            return 'down'

    def start(self, ndata, toolbox_dirs, args):
        """harkmw プロセスを開始する

        Parameters
        ----------
        ndata : str
            ネットワーク定義ファイルのパス
        toolbox_dirs : list
            tlb ファイル等を検索するパスのリスト
        args : object
            MQTT から受信した起動時の引数
        """
        logger.debug('Subprocess: start')

        notify_topic = '{}/res/start/status'.format(self.process_name)
        logger.debug('Subprocess: notify-topic={}'.format(notify_topic))

        pid = os.fork()
        if pid == 0:
            logger.debug('Subprocess: fork child')
            # child
            try:
                app = harkmw.Process(
                    args, broker=self.daemon.broker, topic=self.daemon.topic)
                app.add_toolbox_dirs(toolbox_dirs)
                app.parse(ndata)
                app.run(self.process_name, notify_init=notify_topic)

            except Exception as ex:
                logger.error('Exception in subprocess: {}'.format(ex))
                os._exit(1)

            logger.debug('Subprocess: exit child')
            os._exit(0)

        else:
            logger.debug('Subprocess: fork parent: child_pid={}'.format(pid))
            # parent
            self.child_pid = pid
            return

    def stop(self):
        """子プロセスを停止する
        """
        self.terminate()

    def terminate(self, kill=False):
        """子プロセスを強制停止する
        """
        if self.child_pid > 0:
            if kill:
                os.kill(self.child_pid, signal.SIGKILL)
            else:
                os.kill(self.child_pid, signal.SIGTERM)

    def cleanup(self):
        """子プロセスに関する情報を開放する
        """
        self.child_pid = None


class Daemon():
    """harkmw daemon の処理を管理するクラス
    """

    def __init__(self, broker, topic):
        """harkmw daemon の管理情報を初期化する

        Parameters
        ----------
        broker : str
            MQTT ブローカとして接続するアドレス
        topic : str
            MQTT を通して受信するためのトピック名
        """
        self.broker = broker
        self.topic = topic
        self.process_names = []
        self.toolbox_dirs = []
        self.publish_queue = []

        self.to_stop_service = False
        self.mqtt_loop_interval = 0.1

    def add_toolbox_dirs(self, dirs):
        """HARK tlb を検索するディレクトリを追加する

        Parameters
        ----------
        dirs : list
            ディレクトリ名のリスト
        """
        self.toolbox_dirs.extend(dirs)

    def run(self, process_names):
        """harkmw daemon の初期化、実行等一連の処理を行う

        Parameters
        ----------
        process_names : str
            この daemon が対応するプロセス名のリスト
        """
        logger.info(
            'Start daemon process: process_names={}'.format(process_names))
        print('Thread: {}'.format(threading.get_ident()))
        self.process_names = copy(process_names)
        self.subprocesses = dict(
            [(pn, SubProcess(daemon=self, process_name=pn)) for pn in process_names])

        try:
            # try connect to broker
            self.mosquitto_subscribers = {}
            self.mosquitto_client = mqtt.Client()
            self.mosquitto_client.user_data_set(self)
            self.mosquitto_client.on_connect = self.on_mqtt_connect
            self.mosquitto_client.on_disconnect = self.on_mqtt_disconnect
            self.mosquitto_client.on_message = self.on_mqtt_message
            self.mosquitto_client.on_subscribe = self.on_mqtt_subscribe
            self.mosquitto_client.on_unsubscribe = self.on_mqtt_unsubscribe

            args = {}
            if ':' in self.broker:
                host, port = self.broker.split(':')
                args['port'] = int(port)
            else:
                host = self.broker

            self.mosquitto_client.connect(host, **args)
            self.mosquitto_client.subscribe(
                '{}/_DAEMON/req/#'.format(self.topic), 0)

            while not self.to_stop_service:
                self.mosquitto_client.loop(self.mqtt_loop_interval)
                self.check_subprocesses()
                while len(self.publish_queue) > 0:
                    p = self.publish_queue.pop()
                    self.publish_response(p[0], p[1], 0, p[2])

        except BaseException as ex:
            traceback.print_exc()
            logger.error('Exception in process: {}'.format(ex))

        finally:
            logger.info('End daemon process')

    def push_publish(self, pname, topic, message):
        """MQTTを介して送信する情報をキューに蓄積する

        Parameters
        ----------
        pname : str
            プロセス名
        topic : str
            MQTT トピック名
        message : str
            送信するメッセージ
        """
        logger.debug('push_publish: pname={}, topic={}'.format(pname, topic))
        self.publish_queue.append((pname, topic, message))

    def publish_response(self, pname, topic, count, data):
        """MQTTを介してメッセージを送信する

        Parameters
        ----------
        pname : str
            プロセス名
        topic : str
            トピック名
        count : int
            時刻フレーム値
        data : object
            送信するデータ
        """
        t = '{}/{}/{}'.format(self.topic, pname, topic)
        p = bytes('HARK' + '0001' + 'PCKL', encoding='utf8') + \
            pickle.dumps({'count': count, 'data': data, 'topic': t})
        logger.debug('Publish: topic={}'.format(t))
        result, qid = self.mosquitto_client.publish(t, p, 0)

    def on_mqtt_connect(self, client, userdata, flags, rc):
        """MQTTブローカに接続した場合の処理

        ログ出力以外処理は行わない
        """
        logger.info('MQTT Connected: client={}, flags={}, rc={}'.format(
            client, flags, rc))

    def on_mqtt_disconnect(self, client, userdata, rc):
        """MQTTブローカから切断した場合の処理

        ログ出力以外処理は行わない
        """
        logger.info('MQTT Disconnected: client={}, rc={}'.format(client, rc))

    def on_mqtt_subscribe(self, client, userdata, mid, granted_qos):
        """MQTTブローカに subscribe した場合の処理

        ログ出力以外処理は行わない
        """
        logger.info('MQTT Subscribe: client={}, mid={}, granted_qos={}'.format(
            client, mid, granted_qos))

    def on_mqtt_unsubscribe(self, client, userdata, mid):
        """MQTTブローカに unsubscribe した場合の処理

        ログ出力以外処理は行わない
        """
        logger.info('MQTT Unsubscribe: client={}, mid={}'.format(client, mid))

    def on_mqtt_message(self, mosquitto, obj, msg):
        """MQTTブローカからメッセージを受信した場合の処理

        Parameters
        ----------
        mosquitto : object
            未使用
        obj : object
            未使用
        msg : object
            受信したメッセージ
        """
        message = msg.payload
        topic = msg.topic
        retain = msg.retain
        logger.info(
            'MQTT message: topic={}, retain={}, message-size={}'.format(topic, retain, len(message)))

        topic_base = '{}/_DAEMON'.format(self.topic)
        sub_topic = topic[len(topic_base):].lstrip('/')

        if not sub_topic:
            # topic does not begin with self.topic
            logger.warning('Ignore MQTT message: topic={}'.format(topic))

        else:
            sub_topics = sub_topic.split('/')
            rtype, args = sub_topics[0], sub_topics[1:]

            if rtype == 'req':
                if len(args) == 0:
                    logger.warning(
                        'Received unexpeced request: No command given: topic={}'.format(topic))

                elif len(args) == 1:
                    logger.warning(
                        'Received unexpeced request: No process name given: topic={}'.format(topic))

                else:
                    command, pname, args = args[0], args[1], args[2:]
                    logger.debug(
                        'Received: command={}, pname={}, args={}'.format(
                            command, pname, args))
                    if pname not in self.process_names:
                        logger.info(
                            'Unexpected process name: ignored. process_name={}'.format(pname))

                    else:
                        sp = self.subprocesses[pname]

                        try:
                            body = pickle.loads(message)
                        except pickle.PickleError as ex:
                            logger.warning(
                                'Failed unpickle message: ex={}'.format(ex))

                        if command == 'status':
                            self.push_publish(
                                pname, 'res/{}/status'.format(command), sp.get_status())

                        elif command == 'start':
                            sp = self.subprocesses[pname]
                            if sp.child_pid is None:
                                logger.info(
                                    'Starting process "{}"'.format(pname))
                                sp.start(body['ndata'],
                                         self.toolbox_dirs, body['args'])
                            else:
                                logger.info(
                                    'Received start request for "{}", but already running.'.format(pname))
                                self.push_publish(
                                    pname, 'res/{}/status'.format(command), sp.get_status())

                        elif command == 'stop':
                            sp = self.subprocesses[pname]
                            sp.stop()
                            self.push_publish(
                                pname, 'res/{}/status'.format(command), sp.get_status())

                        elif command == 'terminate':
                            sp = self.subprocesses[pname]
                            sp.terminate()
                            self.push_publish(
                                pname, 'res/{}/status'.format(command), sp.get_status())

                        else:
                            logger.warning(
                                'Received unexpeced request: command={}, topic={}'.format(
                                    command, topic))

            elif rtype in ('res',):
                # ignore res(response) message
                logger.debug(
                    'Ignore res message: type={}: topic={}'.format(
                        rtype, topic))

            elif rtype in ('notify',):
                logger.debug(
                    'Ignore notify message: type={}: topic={}'.format(
                        rtype, topic))

            else:
                logger.warning(
                    'Received unexpeced message: type={}: topic={}'.format(
                        rtype, topic))

    def check_subprocesses(self):
        """子プロセスの状態を検証する
        """
        for spname, sp in self.subprocesses.items():
            if sp and sp.child_pid:
                cpid, status = os.waitpid(sp.child_pid, os.WNOHANG)
                if cpid == 0:
                    continue

                logger.debug(
                    'waitpid for pid {} returned with status: {}.'.format(
                        sp.child_pid, status))

                args = {'cpid': cpid}
                cleanup = False
                if os.WIFCONTINUED(status):
                    message = 'Child process {} was continued.'.format(cpid)
                elif os.WIFSTOPPED(status):
                    message = 'Child process {} was stopped by signal {}.'.format(
                        cpid, os.WSTOPSIG(status))
                    args['stopsig'] = os.WSTOPSIG(status)
                elif os.WIFSIGNALED(status):
                    message = 'Child process {} was termionated by signal {}.'.format(
                        cpid, os.WTERMSIG(status))
                    args['termsig'] = os.WTERMSIG(status)
                    args['dmpcore'] = os.WCOREDUMP(status)
                    cleanup = True
                elif os.WIFEXITED(status):
                    message = 'Child process {} exited normally, with exit code {}.'.format(
                        cpid, os.WEXITSTATUS(status))
                    args['scode'] = os.WEXITSTATUS(status)
                    cleanup = True

                logger.debug('cleaning up.')
                if cleanup:
                    sp.cleanup()

                message = message.format(**args)
                logger.debug('check-status: {}'.format(message))

                body = {
                    'message': message,
                    'status': sp.get_status(),
                }
                self.push_publish(spname, 'nofity/status', body)

# end of file
