#! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
# 
# Licensed under the HARK license, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# https://www.hark.jp/notice/HARK_License_Agreement.pdf

import logging
import logging.config
import os

logger = logging.getLogger('hark-middleware')


def initialize_log(
        config_file=None,
        log_level='info',
        output_file=None,
        output_fd=None):
    """ログ出力を初期化する

    Parameters
    ----------
    config_file : str
        ログ設定ファイル
    log_level : str
        出力するログのレベル
    output_file : str
        ログを出力するファイルのパス
    output_fd : int
        ログを指定した file descriptor に出力する場合に指定する
    """
    global logger

    if config_file:
        logging.config.fileConfig(config_file)
    if output_file is not None:
        setup_default(log_level, logging.FileHandler(output_file))
    if output_fd is not None:
        setup_default(log_level, FDHandler(output_fd))
    if not config_file and output_file is None and output_fd is None:
        setup_default(log_level, logging.StreamHandler())

    logger = logging.getLogger('hark-middleware')

    # # 'application' code
    # logger.debug('debug message')
    # logger.info('info message')
    # logger.warning('warn message')
    # logger.error('error message')
    # logger.critical('critical message')


def setup_default(log_level, ch):
    """既定値を用いてログを初期化する

    Parameters
    ----------
    log_level : str
        ログレベル
    ch : object
        ログハンドラ
    """
    global logger

    log_level = log_level.upper()
    if log_level == 'DEBUG':
        level = logging.DEBUG
    elif log_level == 'INFO':
        level = logging.INFO
    elif log_level in ('WARN', 'WARNING'):
        level = logging.WARNING
    elif log_level == 'ERROR':
        level = logging.ERROR
    elif log_level == 'CRITICAL':
        level = logging.CRITICAL
    else:
        raise BaseException('Unexpected log level value: {}'.format(log_level))

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(process)d - %(levelname)s - %(message)s')
    ch.setLevel(level)
    ch.setFormatter(formatter)

    logger = logging.getLogger('hark-middleware')
    logger.setLevel(level)
    logger.addHandler(ch)


class FDHandler(logging.StreamHandler):
    """ファイルデスクリプタに出力するためのログハンドラ
    """

    def __init__(self, fd, mode='w', encoding=None, delay=False, errors=None):
        """ログハンドラを初期化する

        Parameters
        ----------
        fd : int
            ファイルデスクリプタ値
        mode : str
            アクセス種類。通常 'w' を与える
        endoding : object
            ログ出力に用いるエンコーディング
        delay : bool
            未使用
        errors : object
            未使用
        """
        self.fd = fd
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.delay = delay
        if delay:
            # We don't open the stream, but we still need to call the
            # Handler constructor to set level, formatter, lock etc.
            logging.Handler.__init__(self)
            self.stream = None
        else:
            logging.StreamHandler.__init__(self, self._open())

    def _open(self):
        """指定したファイルデスクリプタを開く
        """
        return os.fdopen(
            self.fd,
            self.mode,
            encoding=self.encoding,
            errors=self.errors)

    def __repr__(self):
        """デバッグ用のダンプ出力
        """
        level = str(self.level)
        return '<%s %d (%s)>' % (self.__class__.__name__, self.fd, level)


# end of file
