#! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
# 
# Licensed under the HARK license, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# https://www.hark.jp/notice/HARK_License_Agreement.pdf

from harkmw.logger import logger
import harkmw.logic.node
from collections import defaultdict
import os

debug = (bool(os.environ['HARK_DEBUG'])
         ) if 'HARK_DEBUG' in os.environ else False


class MQTTNode(harkmw.logic.node.Node):
    """MQTTを介して別プロセスで稼働しているノードを管理するクラス
    """
    def __init__(self):
        """クラスの初期化を行う
        """
        harkmw.logic.node.Node.__init__(self)
        self.inputs = {}
        self.outputs = defaultdict(lambda: [])
        self.data = defaultdict(lambda: {})
        self.inputLookWindows = defaultdict(lambda: (0, 0))
        self.outputLookWindows = defaultdict(lambda: (0, 0))
        self.network = None

    def build(self, network, node_def, params):
        """ノードを構築する

        ノードの実体は別プロセスにあるため、C++ で実装された HARK ノードを扱うことはなく、
        管理用の情報を保持するのみである

        Parameters
        ----------
        network : Network
            このノードが属するネットワーク
        node_def : NodeDef
            ノードの定義情報
        params : Parameters
            起動時にユーザが指定した値も含んだ、ネットワークパラメータ情報の集合
        """
        self.network = network
        self.type = node_def.type
        self.name = node_def.name
        self.topic = node_def.topic
        self.process_name = node_def.get_process_name()

    def process(self, output, count):
        """ノードの処理を実行する

        ノードの実体は別プロセスにあるため実際には何も行わない
        """
        pass

    def get_output(self, output_name, count):
        """ノードの出力値を取得する

        Parameters
        ----------
        output_name : str
            出力端子名
        count : int
            取得の対象となるフレーム番号
        """
        if output_name in self.data:
            output_data = self.data[output_name]
            if count in output_data:
                return output_data[count]
            else:
                raise BaseException(
                    'MQTTNode.get_output: no data with count {} in data with name {}'.format(
                        count, output_name))

        else:
            raise BaseException(
                'MQTTNode.get_output: no data with name {}, type={}, name={}'.format(
                    output_name, self.type, self.name))

    def get_input_count(self):
        """入力端子の個数を取得する
        """
        return len(self.inputs)

    def get_output_count(self):
        """出力端子の個数を取得する
        """
        return len(self.outputs)

    def get_input_look_window(self, input_id):
        """HARK ノードの入力端子から lookAhead, lookBack を取得する

        Parameters
        ----------
        input_id : str
            入力端子名
        """
        return self.inputLookWindows[input_id]

    def set_input_look_window(self, input_id, lookWindow):
        """HARK ノードの入力端子に lookAhead, lookBack を設定する

        Parameters
        ----------
        input_id : str
            入力端子名
        look_window : tuple
            lookAhead, lookBack の 2 値からなる tuple
        """
        self.inputLookWindows[input_id] = lookWindow

    def get_output_look_window(self, output_name):
        """HARK ノードの出力端子から lookAhead, lookBack を取得する

        Parameters
        ----------
        output_id : str
            出力端子名
        """
        return self.outputLookWindows[output_name]

    def set_output_look_window(self, output_id, lookWindow):
        """HARK ノードの出力端子に lookAhead, lookBack を設定する

        Parameters
        ----------
        output_id : str
            出力端子名
        look_window : tuple
            lookAhead, lookBack の 2 値からなる tuple
        """
        self.outputLookWindows[output_id] = lookWindow

    def connect_node(self, terminal_name, node_from, terminal_from_name):
        """指定した HARK ノードと接続する

        ノードの実体は別プロセスにあるため実際には何も行わない

        Parameters
        ----------
        terminal_name : str
            未使用
        node_from : NativeNode
            未使用
        terminal_from_name : str
            未使用
        """
        pass

    def push_data(self, input, count, process, internal=False, outer_count=-1):
        """ノードにデータを送る

        Parameters
        ----------
        inpput : object
            未使用
        count : int
            処理の対象となるフレーム番号
        process : Process
            harkmw プロセス
        internal : object
            未使用
        outer_count : object
            未使用
        """
        logger.debug(
            "MQTTNode.push_data: Try push data for Node [{}/{}], count [{}/{}], #inputs={}, #outputs={}, ({})".format(
                self.type, self.name, count, self.count, len(
                    self.inputs), len(
                    self.outputs), self))

        # if False:
        #     process.publish(self.get_topic() + "/_INPUT", count, True)
        #
        # else:
        #     process.subscribe(self.get_topic() + "/_INPUT/ack", self)
        #     process.publish(self.get_topic() + "/_INPUT", count, True)
        #
        #     # wait for response
        #     process.wait_topic(self.get_topic() + "/_INPUT/ack")
        #     process.unsubscribe(self.get_topic() + "/_INPUT/ack", self)

        if len(self.inputs) == 0:
            t = '{}/_INPUT'.format(self.get_topic())
            tack = '{}/ack'.format(t)
            terr = '{}/err'.format(t)
            process.subscribe([tack, terr], None)
            process.publish(t, count, True)

            # wait for response
            try:
                topic, data = process.wait_topic([tack, terr])
                if topic == tack:
                    pass
                elif topic == terr:
                    logger.error('Remote process returned error: ')
                    for d in data["data"]:
                        for i, l in enumerate(str(d).splitlines()):
                            logger.error(
                                '- {}{}'.format(("- " if i > 0 else ""), l))

                    raise Exception(
                        'Remote process returned error: {}'.format(
                            data["data"]))

            finally:
                process.unsubscribe([tack, terr], None)

        # logger.debug("MQTTNode.push_data: return: Node [{}/{}], count
        # [{}/{}], #inputs={}, #outputs={}, ({})".format(self.type, self.name,
        # count, self.count, len(self.inputs), len(self.outputs), self))
        return count

    def dump(self):
        """デバッグ用のダンプ出力
        """
        print('[MQTTNode]')
        print('  type: %s' % self.type)
        print('  name: %s' % self.name)
        print('  network: type=%s, name=%s' %
              (self.network.type, self.network.name))

        print('  inputs: ')
        for (name, terminal) in self.inputs.items():
            print(
                '    Input: name=%s, node=%s/%s, terminal=%s' %
                (name,
                 terminal.input_node.type,
                 terminal.input_node.name,
                 terminal.input_terminal_name))

        print('  outputs: ')
        for (name, proxy_list) in self.outputs.items():
            for proxy in proxy_list:
                print(
                    '    Output: name=%s, node=%s/%s, terminal=%s' %
                    (name,
                     proxy.output_node.type,
                     proxy.output_node.name,
                     proxy.output_terminal_name))

    def __str__(self):
        """デバッグ用のダンプ出力
        """
        return "<%s: type=%s, name=%s>" % (
            self.__class__.__name__, self.type, self.name)

# end of file
