/* -*- mode: C++; coding: utf-8 -*- */

/*
 * Copyright (c) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "libharkmw.h"
#include "node.h"

#include <hark-base/object_param.h>
#include <hark-base/Exception.h>

#include "type.h"
#include "collector.h"

namespace py = pybind11;

namespace HARKMW
{
Node::~Node()
{
    TRY_HARK(delete this->node, "Node deletion");
}

Node::Node(py::object node, const std::string &name, py::object parameters) :
    node(NULL)
{
    HARK::_NodeFactory *factory = nullptr;
    TRY_HARK(factory = HARK::Node::getFactoryNamed(name), "finding node factory");
    if(!factory) {
        std::stringstream ss;
        ss << "No node class with name '" << name << "' found";
        throw HARKMW::node_not_implemented(ss.str());
    }

    py::object paramsObject = parameters.attr("params");
    std::map<std::string, py::object> params = py::cast<std::map<std::string, py::object>>(paramsObject);

    HARK::ParameterSet pset;
    for(const auto &p: params) {
        const std::string &name  = p.first;
        const std::string &type  = py::cast<const std::string>(py::object(p.second.attr("type")));
        const std::string &value = py::cast<const std::string>(py::object(p.second.attr("value")));
        std::cout << "name=" << name << ", type=" << type << ", value=" << value << std::endl;
        if(value.empty()) {
            continue;
        }
        pset.add(name, HARK::ObjectParam::stringParam(type, value, pset));
    }

    /* Do processes */
    TRY_HARK(this->node = factory->Create(name, pset), "node creation");
    this->nodeObject = node;
}

void
Node::process(const std::string &output_name, int count, int input_count, int output_count)
{
    HARK::ObjectRef output;
    if(this->node->getName() == "HARKMW_Collector") {
        HARK::HARKMW_Collector *collector = nullptr;
        TRY_HARK(collector = dynamic_cast<HARK::HARKMW_Collector *>(this->node), "node casting");
        TRY_HARK(output = collector->getOutputNamed(output_name, input_count, output_count), "getting output from collector");
    } else {
        TRY_HARK(output = this->node->getOutputNamed(output_name, count), "getting output from node");
    }
}

py::object
Node::get_output(const std::string &name, int count)
{
    HARK::ObjectRef o;
    TRY_HARK(o = this->node->getOutputNamed(name, count), "getting output from node");
    return toPyObject(o);
}

void
Node::connect(const std::string &in, Node *inputNode, const std::string &out)
{
    TRY_HARK(this->node->connectToNode(in, inputNode->node, out), "connecting nodes");
}

void
Node::initialize()
{
    TRY_HARK(this->node->initialize(), "during initialize() call");
}

size_t
Node::get_input_count() const
{
    return this->node->inputs.size();
}

size_t
Node::get_output_count() const
{
    return this->node->outputNames.size();
}


std::pair<int, int>
Node::get_output_look_window(const std::string &output_id) const
{
    HARK::BufferedNode *bufferedNode;
    TRY_HARK(bufferedNode = dynamic_cast<HARK::BufferedNode *>(this->node), "node casting");

    int look_ahead = 0, look_back = 0;
    if(bufferedNode) {
        int id;
        TRY_HARK(id = bufferedNode->getOutputID(output_id), "retrieving terminal id from name");
        look_ahead = bufferedNode->outputs[id].lookAhead;
        look_back  = bufferedNode->outputs[id].lookBack;
    }

    return std::pair<int, int>(look_ahead, look_back);
}

void
Node::set_output_look_window(const std::string &output_id, int lookAhead, int lookBack)
{
    HARK::BufferedNode *bufferedNode;
    TRY_HARK(bufferedNode = dynamic_cast<HARK::BufferedNode *>(this->node), "node casting");
    if(bufferedNode) {
        int id;
        TRY_HARK(id = bufferedNode->getOutputID(output_id), "retrieving terminal id from name");
        bufferedNode->outputs[id].lookAhead = lookAhead;
        bufferedNode->outputs[id].lookBack  = lookBack;
    }
}

std::pair<int, int>
Node::get_input_look_window(const std::string &input_id) const
{
    HARK::BufferedNode *bufferedNode;
    TRY_HARK(bufferedNode = dynamic_cast<HARK::BufferedNode *>(this->node), "node casting");
    int look_ahead = 0, look_back = 0;
    if(bufferedNode) {
        int id;
        TRY_HARK(id = bufferedNode->getInputID(input_id), "retrieving terminal id from name");
        look_ahead = bufferedNode->inputsCache[id].lookAhead;
        look_back  = bufferedNode->inputsCache[id].lookBack;
    }

    return std::pair<int, int>(look_ahead, look_back);
}

void
Node::set_input_look_window(const std::string &input_id, int lookAhead, int lookBack)
{
    HARK::BufferedNode *bufferedNode;
    TRY_HARK(bufferedNode = dynamic_cast<HARK::BufferedNode *>(this->node), "node casting");
    if(bufferedNode) {
        int id;
        TRY_HARK(id = bufferedNode->getInputID(input_id), "retrieving terminal id from name");
        bufferedNode->inputsCache[id].lookAhead = lookAhead;
        bufferedNode->inputsCache[id].lookBack  = lookBack;
    }
}

void
Node::dump_input(int count) const
{
    for(size_t i = 0; i < this->node->inputs.size(); i++) {
        HARK::ObjectRef o;
        TRY_HARK(o = this->node->getInput(i, count), "input node retrieval");
        TRY_HARK(o->printOn(), "dump input");
        std::cout << std::endl;
    }
}
}

/* end of file */
