/* -*- mode: C++; coding: utf-8 -*- */

/*
 * Copyright (c) 2013 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "libharkmw.h"
#include "proxy.h"

#include <hark-base/BufferedNode.h>
#include <hark-base/Collector.h>
#include <hark-base/object_param.h>
#include <hark-base/Vector.h>

#include "node.h"
#include "type.h"

#define PROXY_USING_BUFFER_NODE 0
#define VERBOSE_LOG 0

namespace HARKMW
{
class ProxyNode : public HARK::Node
{
private:
    int count;
    Proxy *proxy;

    std::vector<HARK::ObjectRef> buffer;
    size_t position;
    size_t queuemax;

    //for HARK use
    int input_id;
    int output_id;

public:
    std::string tag;
    size_t lookAhead, lookBack;

public:
    ProxyNode(std::string nodeName, const HARK::ParameterSet &params, Proxy *proxy) :
        HARK::Node(nodeName, params)
    {
        this->count     = -1;
        this->input_id  = this->addInput ("I/O");
        this->output_id = this->addOutput("I/O");
        this->proxy = proxy;

        this->buffer.clear();
        this->position = 0;
        this->queuemax = 0;
        this->lookAhead = 0;
        this->lookBack = 0;
    }

public:
    void force_count(int count)
    {
        this->count = count;
    }

    HARK::ObjectRef calculate(int count)
    {
#if VERBOSE_LOG
        std::cout << "calculate: ProxyNode[" << this->tag << "]: count=" << count << std::endl;
#endif
        HARK::NodeInput &input = inputs[this->input_id];
        HARK::ObjectRef inputValue;
        if(input.node) {
            inputValue = input.node->getOutput(input.outputID, count);
        } else {
            py::object input_node = this->proxy->proxyObject.attr("input_node");
            py::object input_terminal_name = this->proxy->proxyObject.attr("input_terminal_name");
            py::object output_value = input_node.attr("get_output")(input_terminal_name, count);

            inputValue = fromPyObject(output_value);
        }

        return inputValue;
    }

    virtual HARK::ObjectRef getOutput(int output_id, int count)
    {
        return this->getOutput(output_id, count, false);
    }
    virtual HARK::ObjectRef getOutput(int output_id, int count, bool no_remove)
    {
#if VERBOSE_LOG
        std::cout << "ProxyNode[" << this->tag << "]::getOutput: output_id=" << output_id << ", count=" << count << ", lookAhead=" << this->lookAhead << ", lookBack=" << this->lookBack << std::endl;
#endif
        if((size_t)count < this->position) {
            std::stringstream ss;
            ss << "count " << count << " is out of bounds: [" << this->position << ", " << this->position + this->buffer.size() << "], tag=" << this->tag;
            throw new HARK::NodeException(this, ss.str().c_str(), __FILE__, __LINE__);

        } else {
            while(this->position + this->buffer.size() < (size_t)count + 1) {
                this->buffer.push_back(this->calculate(this->position + this->buffer.size()));
            }
        }

        HARK::ObjectRef outValue = this->buffer[count - this->position];

        if(no_remove) {
            if(this->queuemax < this->buffer.size()) {
                this->queuemax = this->buffer.size();
            }

        } else {
            if(this->lookAhead + this->lookBack + 1 < this->buffer.size() && this->queuemax < this->buffer.size() + 1) {
                int to_remove = 1;
#if VERBOSE_LOG
                std::cout << "ProxyNode[" << this->tag << "]::getOutput: remove " << to_remove << " elements." << std::endl;
#endif
                this->buffer.erase(this->buffer.begin(), this->buffer.begin() + to_remove);
                this->position += to_remove;
            }
        }
#if VERBOSE_LOG
        std::cout << "ProxyNode[" << this->tag << "]::getOutput: buffer-size: " << this->buffer.size() << std::endl;
#endif
        return outValue;
    }

    void setOutput(int output_id, int count, HARK::ObjectRef &data)
    {
        while(this->position + this->buffer.size() < (size_t)count) {
            this->buffer.push_back(this->calculate(this->position + this->buffer.size()));
        }
        this->buffer.push_back(data);
        if(this->queuemax < this->buffer.size()) {
            this->queuemax = this->buffer.size();
        }
    }

    virtual void initialize()
    {
        this->HARK::Node::initialize();
    }

};

Proxy::Proxy(py::object proxy)
{
    HARK::ParameterSet emptyParams;
    this->proxyObject = proxy;
    this->proxyNode   = new ProxyNode(std::string("ProxyNode"), emptyParams, this);
}

Proxy::~Proxy()
{
    delete this->proxyNode;
}

void
Proxy::connect_input(py::object node, const std::string &terminal_name)
{
    HARKMW::Node *harkmw_node = py::cast<HARKMW::Node *>(node);
    TRY_HARK(this->proxyNode->connectToNode(std::string("I/O"), harkmw_node->node, terminal_name), "node connection");
}

void
Proxy::connect_output(py::object node, const std::string &terminal_name)
{
    HARKMW::Node *harkmw_node = py::cast<HARKMW::Node *>(node);
    HARK::Node *_node = harkmw_node->node;
    HARK::BufferedNode *bufferedNode;
    TRY_HARK(bufferedNode = dynamic_cast<HARK::BufferedNode *>(_node), "node casting");
    if(bufferedNode) {
        std::string name = terminal_name;
        int inputId = -1;
        for(std::vector<HARK::NodeInput>::const_iterator it = _node->inputs.begin(); it != _node->inputs.end(); ++it) {
            if((*it).name == name) {
                inputId = (it - _node->inputs.begin());
                break;
            }
        }
        if(inputId >= 0) {
            this->proxyNode->lookAhead = bufferedNode->inputsCache[inputId].lookAhead;
            this->proxyNode->lookBack = bufferedNode->inputsCache[inputId].lookBack;
        }
    }

    TRY_HARK(_node->connectToNode(terminal_name, this->proxyNode, std::string("I/O")), "node connection");
}

void
Proxy::initialize(const std::string &tag)
{
    this->proxyNode->tag = tag;
    TRY_HARK(this->proxyNode->initialize(), "initialization");
}

int
Proxy::push_data(int count)
{
    TRY_HARK(this->proxyNode->getOutput(0, count, true), "data pushing");
    return count;
}

py::object
Proxy::get_data(int count)
{
    HARK::ObjectRef out;
    TRY_HARK(out = this->proxyNode->getOutput(0, count), "retrieving data");
    return toPyObject(out);
}

void
Proxy::set_data(int count, py::object data)
{
    HARK::ObjectRef value = fromPyObject(data);
    TRY_HARK(this->proxyNode->setOutput(0, count, value), "data assignment");
}

std::pair<int, int>
Proxy::get_look_window()
{
    return std::pair<int, int>(this->proxyNode->lookAhead, this->proxyNode->lookBack);
}

void
Proxy::set_look_window(int lookAhead, int lookBack)
{
    this->proxyNode->lookAhead = lookAhead;
    this->proxyNode->lookBack = lookBack;
}

void
Proxy::force_count(int count)
{
    this->proxyNode->force_count(count);
}

}

/* end of file */
